/*
 * Log4JFugue - Application Sonification
 * Copyright (C) 2008-2009  Brian Tarbox
 *
 * http://www.log4jfugue.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

package org.log4jfugue;
import java.util.*;
import java.io.Reader;
import java.io.BufferedReader;
import org.apache.log4j.Logger;
import com.eaio.stringsearch.*;
import com.eaio.stringsearch.BoyerMooreHorspool;

/**
 * Listens for log messages to process.  This is the class that does the string searches
 * to determine that a particular message of interest has been seen.  When this happens the
 * count of the associated instrument is incremented.  The message can arrive from a variety of sources 
 * such as a live feed or a replay of an existing log file.  The actual source does not
 * matter to this class.
 * 
 * A recent change to this class has been to replace the native java string search with an
 * MIT based BoyerMoorHorspool string search that is up to 10x faster.
 * 
 * @author btarbox
 *
 */
public class MessageReceiver extends Thread {
	Logger log = Logger.getLogger(MessageReceiver.class);
	private HashMap<String, String> instrumentMessageMap = new HashMap<String, String>();

	// use the MIT fast string search algorithm with pre-processed patterns
	private HashMap<String, Object> stringSearchMap = new HashMap<String, Object>();
	private StringSearch searcher = new BoyerMooreHorspool();
	
	private Music music = null;
    private int counterIndex = 0;
    private Properties messages;
    private BufferedReader reader = null;
    private Statistics statistics = null;
    
    public MessageReceiver() {
    	log.info("at MessageReceiver constructor");
    }

    /**
     * Used by Spring DI to tell class where to get the message Reader
     * @param reader
     */
    public void setReader(Reader reader) {
    	log.info("set buffered reader");
    	this.reader = new BufferedReader(reader);
    }
    
    public void setStatistics(Statistics statistics) {
    	this.statistics = statistics;
    }
    
    /** 
     * Used by Spring DI to tell class what the main Music object is
     * @param music
     */
    public void setMusic(Music music) {
    	log.info("set music");
        this.music = music;
     }

    /**
     * Injected by Spring, after music is set.  The properties object contains
     * the definitions of the strings to look for and the instruments to 
     * associate with them
     * @param messages
     */
    public void setMessages(Properties messages) {
       log.info("setMessages");
       this.messages = messages;
       log.info("the messages are " + messages.toString());
       try {
    	   initializeMessagemap(messages);
       }catch(Exception ex) {
    	   log.error("exception initializingMessageMap ", ex);
       }
       try {
    	   music.setProperties(messages);
       }catch(Exception ex) {
    	   log.error("exception setting messages to music");
       }
    }
    
    private void initializeMessagemap(Properties messages) {
    	Enumeration e = messages.propertyNames();
    	while(e.hasMoreElements()) {
    		String key = (String)e.nextElement();
    		String pattern = messages.getProperty(key);
    		log.info("putting " + key + ":" + pattern + " into message map");
    		this.instrumentMessageMap.put(key, pattern);
    		
    		Object processedPattern = searcher.processString(pattern);
    		this.stringSearchMap.put(key, processedPattern);
    	}
    }
    
    /**
     * Let other classes, like the RhythmSoundBuilder get our list of instruments
     */
    public Set<String> getInstrumentList() {
    	return this.instrumentMessageMap.keySet();
    }
    
    /**
     * There are two message accumulators, one is being added to and the other is
     * being played from.  This method allows the player to swap the accumulators
     */
    public void swapCounters() {
        counterIndex = counterIndex==0 ? 1 : 0;
    }

    /**
     * this takes a line of text (from wherever) and sees if it matches any
     * of our target string.  If the system has any processing bottleneck
     * it would be here
     * @param searchMe
     * @deprecated   use processAMessageFaster instead
     */
    private void processAMessage(String searchMe) {
    	if(log.isDebugEnabled())
    		log.debug("processing the message " + searchMe + ".");
    	Iterator it = instrumentMessageMap.keySet().iterator();
    	while(it.hasNext()) {
    		String key = (String)it.next();
    		String searchString = this.instrumentMessageMap.get(key);
        	if(log.isDebugEnabled())
        		log.debug("testing for message " + searchString);
    		if(searchMe.indexOf(searchString) != -1) {
    	    	if(log.isDebugEnabled())
    	    		log.debug("found an instance of text for instrument " + key);
    			music.incrementInstrumentCount(key);
    		}
    	}
    }
    

    private void processAMessageFaster(String searchMe) {
    	if(log.isDebugEnabled())
    		log.debug("processing the message " + searchMe + ".");
    	Iterator it = instrumentMessageMap.keySet().iterator();
    	while(it.hasNext()) {
    		String key = (String)it.next();
    		String pattern          = this.instrumentMessageMap.get(key);
    		Object processedPattern = this.stringSearchMap.get(key);
    		if(searcher.searchString(searchMe, pattern, processedPattern) != -1) {
    	    	if(log.isDebugEnabled())
    	    		log.debug("found an instance of text for instrument " + key);
    			music.incrementInstrumentCount(key);
    		}
    	}
    }
    
    
   /**
    * Build the music string to send to JFugue
    */
   public void run() {
	   String line;
	   log.info("at MessageReceiver thread.run");
	   try {
		   while(true) {
		    	if(log.isDebugEnabled())
		    		log.debug("about to wait for line");
		    	
			   line = reader.readLine();
			   
		       if(log.isDebugEnabled())
		    		log.debug("got line " + line);
			   if(line == null)
				   break;
			   
			   processAMessageFaster(line);
			   
			   if(null != statistics) {
				   statistics.processStats(line);
			   }
		   }
	   }catch(Exception ex) {
		   log.error("exception in thread.run ", ex);
	   }
	   log.info("message receiver run loop exiting");
   }
   
}

