/*
 * Log4JFugue - Application Sonification
 * Copyright (C) 2008-2009  Brian Tarbox
 *
 * http://www.log4jfugue.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

package org.log4jfugue;
import java.util.*;

import org.apache.log4j.Logger;
import org.jfugue.Player;
import org.jfugue.Pattern;

/**
 * One of the two primary threads of the system, this class builds and plays
 * the actual music strings.
 * @TODO Should use JFugue's numeric note duration style in buildTempoString
 * 
 * @author btarbox
 *
 */
public class SoundBuilder extends Thread {
	Logger log = Logger.getLogger(SoundBuilder.class);
    protected Music music;
    protected com.sun.media.sound.SoftSynthesizer synth = null;
    protected Player player = null;
    protected SimpleDataGetter dataGetter = null;
    
    public SoundBuilder() {    	
    }
 
    
    /** likely set by spring
     * 
     * @param music
     */
    public void setMusic (Music music) {
    	if(null == music) {
    		throw new NullPointerException("invalid configuration, null music");
    	}
    	this.music = music;
    }

    
    /**
     * set by spring
     * @param player
     */
    public void setPlayer(Player player) {
    	if(null == player) {
    		throw new NullPointerException("invalid configuration, null player");
    	}
    	log.info("set the player class to " + player.getClass().getName());
    	this.player = player;
    }

    public void setDataGetter(SimpleDataGetter dg) {
    	dataGetter = dg;
    }
    
    /**
     * Build the music string to send to JFugue.
     * On each iteration it swaps the accumulator indexes, then starts
     * converting the music counts into music strings.  Once the music
     * strings are built they are sent to JFugue.  The JFugue call blocks 
     * for the duration of the play, which is the only delay in the main
     * for-ever loop of this method.
     */
    public void run() {
    	int index=0;
    	log.info("at SoundBuilder thread.run");
		this.setPriority(Thread.MAX_PRIORITY);
		Pattern pattern = null;
    	final String theBeat = " V9 [BASS_DRUM]w ";
    	try {
    		while(true) {
    			String allVoices = new String(theBeat);
    			music.swapIndex();
    			dataGetter.swappedIndex();
    			allVoices = buildMusicString(allVoices);
    			log.info("about to play: " + index + " " + allVoices);
    			player.play(allVoices); //returns immediately?
    			//player.close();  // why???
    			log.debug("back from play: " + index);
    			index++;
    		}
    	}catch(Exception ex) {
    		log.error("run method threw exception ", ex);
    	}
    }


	protected String buildMusicString(String allVoices) {
		for(int voiceNum =0; voiceNum < Music.MAX_MIDI_VOICES; voiceNum++){
			InstrumentVoice voice = music.getInstrumentVoice(voiceNum);
			if(null != voice) {
				String instrumentName = voice.getInstrumentName();
				Integer voiceNumber   = voice.getVoiceNumber();
				Integer count         = music.getPlayCount(voiceNum);
				String thisVoiceString = buildVoiceString(instrumentName, voiceNumber, count);
				allVoices += thisVoiceString;
			}
		}
		return allVoices;
	}
 
    
    /**
     * Build a string formatted as JFugue needs it.
     * @param instrumentName
     * @param voiceNumber
     * @param count
     * @return
     */
    private String buildVoiceString(String instrumentName, Integer voiceNumber, Integer count) {
    	String thisVoice = null;
    	if(count > 0) {
    	   thisVoice = String.format("V%d I[%s] %s ", voiceNumber, instrumentName, buildTempoString(voiceNumber, count));
    	} else {
    		thisVoice = " ";
    	}
    	log.info("play count for " + instrumentName + " is " + count);
    	return(thisVoice);
    }
    
    
    /**
     * Not sure yet how to write this.  If we have count=4 we want 4 quarter notes.
     * If we have count=12 we want 12 12th notes.  
     * 
     * @param count
     * @return
     */
    private String buildTempoString(Integer voiceNumber, Integer count) {
    	String note;
    	String tempo = "w";
    	if(count > 1) {
    		tempo = "q";
    	} else if(count > 0) {
    		tempo = "h";
    	}
    	
    	switch(voiceNumber) {
    		case 0: note = "A6"; break;
    		case 1: note = "B6"; break;
    		case 2: note = "C6"; break;
    		default: note = "F6";
    	}
    	String thisString = "";
    	for(int x = 0; x < count; x++) {
    		thisString += note + tempo + " ";
    	}
    	return(thisString);
    }
}